
.gajim-incoming-nickname {
    color: rgb(200, 40, 40);
}
.gajim-outgoing-nickname {
    color: rgb(50, 90, 140);
}
.gajim-conversation-text {
    color: inherit;
}
.gajim-status-message {
    color: rgb(100, 130, 160);
}
.gajim-command-error {
    background-color: rgb(10, 50, 70);
    color: rgb(220, 80, 80);
}
.gajim-command-output {
    background-color: rgb(10, 50, 70);
    color: rgb(170, 180, 200);
}
.gajim-user-disconnected {
    color: rgb(230, 170, 170);
}
.gajim-user-connected {
    color: rgb(160, 200, 170);
}
.gajim-status-online {
    color: rgb(102, 191, 16);
}
.gajim-status-away {
    color: rgb(255, 133, 51);
}
.gajim-status-dnd {
    color: rgb(230, 46, 0);
}
.gajim-status-offline {
    color: rgb(154, 154, 154);
}
.gajim-status-connecting {
    color: rgb(77, 166, 255);
}
.gajim-single-emoji-msg {
    font-size: 36pt;
}
.gajim-emoji-msg {
    font-size: 24pt;
}
